import numpy as np
from abc import ABC, abstractmethod


class LeafFunction(ABC):
    """
    Base class leaf value assignment functions
    """
    def __init__(self) -> None:
        self.name = self.__class__.__name__

    def __call__(self, rawPrediction) -> np.ndarray:
        """
        calls '_leafFunc', makes working with this class easier
        """
        return self._leafFunc(rawPrediction)

    @abstractmethod
    def _leafFunc(self, rawPrediction) -> np.ndarray:
        """
        needs to be implemented with every daughter class
        """
        pass


class Mode(LeafFunction):
    """
    This leaf function is used for classification tasks.
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, rawPrediction) -> np.ndarray:
        # If the rawPrediction is one-hot encoded, convert it to categorical
        if len(rawPrediction.shape) == 2 and rawPrediction.shape[1] > 1:
            targetsCategorical = np.argmax(rawPrediction, axis=1)
        else:
            targetsCategorical = rawPrediction

        # Use np.bincount to get the mode
        mode = np.argmax(np.bincount(targetsCategorical.astype(int)))

        return mode


class Mean(LeafFunction):
    """
    this leaf function is used for regression tasks
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, rawPrediction) -> np.ndarray:
        return np.mean(rawPrediction)


class Median(LeafFunction):
    """
    this leaf function is used for regression tasks
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, rawPrediction) -> np.ndarray:
        return np.median(rawPrediction)


class Probabilities(LeafFunction):
    """
    This leaf function is used for classification tasks where you want
    probabilities for each class instead of predicting a single class.
    """
    def __init__(self, numClasses: int) -> None:
        super().__init__()
        self.numClasses = numClasses

    def _leafFunc(self, rawPrediction) -> np.ndarray:
        # If rawPrediction is one-hot encoded, convert it to categorical
        if len(rawPrediction.shape) == 2 and rawPrediction.shape[1] > 1:
            targetsCategorical = np.argmax(rawPrediction, axis=1)
        else:
            targetsCategorical = rawPrediction

        # Initialize a vector of zeros with length equal to the number of classes
        classProbs = np.zeros(self.numClasses)

        # Use advanced integer indexing to fill the appropriate cell
        classProbs[targetsCategorical.astype(int)] = 1

        # Normalize to turn counts into probabilities
        classProbs /= np.sum(classProbs)

        return classProbs


class Confidence(LeafFunction):
    """
    This leaf function is used for classification tasks where you want
    the confidence of the mode class.
    """
    def __init__(self) -> None:
        super().__init__()

    def _leafFunc(self, rawPrediction) -> np.ndarray:
        # Convert one-hot encoded rawPrediction to categorical if applicable
        if len(rawPrediction.shape) == 2 and rawPrediction.shape[1] > 1:
            targetsCategorical = np.argmax(rawPrediction, axis=1)
        else:
            targetsCategorical = rawPrediction.astype(int)

        # Compute mode and its count
        unique, counts = np.unique(targetsCategorical, return_counts=True)
        modeCounts = counts.max()
        modeIndex = np.argmax(counts)
        modeValue = unique[modeIndex]

        # Calculate confidence
        confidence = modeCounts / len(targetsCategorical)

        return np.array([modeValue, modeCounts, confidence])
