import numpy as np
from numpy.typing import ArrayLike
from .decisionTree import DecisionTree
from .voting import Voting, Majority
from .boosting import Boosting, GradientBoosting
from collections import namedtuple
from importlib import import_module


def makeBootstraps(data: ArrayLike, targets: ArrayLike, numBootstraps: int = 100) -> dict:
    """
    This function generates bootstraps for training the tree
    """
    # initialize output dictionary & unique value count
    bootStraps = {}
    uniqueValueCount = 0
    sampleSize = len(data)

    # get list of row indexes
    idx = [i for i in range(sampleSize)]

    # loop through the required number of bootstraps
    for b in range(numBootstraps):
        # obtain boostrap samples with replacement
        sidx = np.random.choice(idx, replace=True, size=sampleSize)
        bootStrapSampleData = data[sidx]
        bootStrapSampleTargets = targets[sidx]

        # compute number of unique values contained in the bootstrap sample
        uniqueValueCount += len(set(sidx))

        # obtain out-of-bag samples for the current b
        oidx = list(set(idx) - set(sidx))
        outOfBagSampleData = np.array([])
        outOfBagSampleTargets = np.array([])
        if oidx:
            outOfBagSampleData = data[oidx]
            outOfBagSampleTargets = targets[oidx]

        # store results
        bootStraps['boot_'+str(b)] = {'boot': [bootStrapSampleData, bootStrapSampleTargets], 'test': [outOfBagSampleData, outOfBagSampleTargets]}

    return bootStraps


class RandomForest(object):
    """
    the random forrest class
    it works like a list for tree
    """
    def __init__(self, *, bootstrapping: bool = True, retrainFirst: bool = False) -> None:
        self.name = self.__class__.__name__
        self.trees = []
        self.bootstrapping = bootstrapping
        self._boosting = False
        self._booster = GradientBoosting()
        self.retrainFirst = retrainFirst

        # the voting algorithm to make the final decision
        self.voting = Majority()

    @property
    def qualifiedName(self) -> tuple:
        return self.__class__.__module__, self.__class__.__name__

    def toDict(self) -> dict:
        saveDict = {}
        saveDict['bootstrapping'] = self.bootstrapping
        saveDict['boosting'] = self._boosting
        saveDict['booster'] = self._booster.name
        saveDict['voting'] = self.voting.name
        saveDict['votingsWeights'] = self.voting.weights.tolist()
        saveDict['trees'] = {}
        for tree in self.trees:
            saveDict['trees'][tree.id] = tree.toDict()
        return saveDict

    @classmethod
    def fromDict(cls, loadDict) -> None:
        instance = cls()  # replace YourClass with the actual name of your class
        instance.bootstrapping = loadDict['bootstrapping']

        Module = import_module('machineLearning.rf.voting')  # dynamically import module
        Class = getattr(Module, loadDict['voting'])  # get class from imported module
        instance.voting = Class(loadDict['votingsWeights'])

        instance._boosting = loadDict['boosting']
        Module = import_module('machineLearning.rf.boosting')  # dynamically import module
        Class = getattr(Module, loadDict['booster'])  # get class from imported module
        instance._booster = Class()

        for id in loadDict['trees']:
            tree = DecisionTree()
            tree = tree.fromDict(loadDict['trees'][id])
            instance.trees.append(tree)
        return instance

    def append(self, tree: DecisionTree) -> None:
        """
        append trees to the class
        """
        assert issubclass(type(tree), DecisionTree), 'only Decision Trees are allowed'
        self.trees.append(tree)

    def setComponent(self, component: Voting | Boosting) -> None:
        """
        setting the voting algorithm
        """
        if isinstance(component, Voting):
            self.voting = component
        elif isinstance(component, Boosting):
            self._booster = component
            self._boosting = True
        else:
            raise TypeError("The given component is not a valid type")

    @property
    def numTrees(self) -> int:
        """
        reutrn the number of tree
        """
        return len(self.trees)

    def train(self, data: ArrayLike, targets: ArrayLike) -> dict:
        """
        protected function to train the ensemble
        """
        self.totalData, self.numFeatures = data.shape
        if self._boosting is False:
           self.retrainFirst = False

        if self.retrainFirst is True:
            treeDict = self.trees[0].toDict()
            tree = DecisionTree.fromDict(treeDict)
            self.trees.append(tree)

        # make bootstrap samples
        if self.bootstrapping:
            bootStraps = makeBootstraps(data, targets, self.numTrees)
        else:
            bootStraps = {}
            for i in range(self.numTrees):
                bootStraps['boot_'+str(i)] = {}
                bootStraps['boot_'+str(i)] = {'boot': [data, targets], 'test': [np.array([]), np.array([])]}

        # iterate through each bootstrap sample & fit a model
        outOfBag = {}
        for i, (strap, tree) in enumerate(zip(bootStraps, self.trees)):
            # skip already trained trees
            if tree._trained is True:
                continue

            if self.bootstrapping:
                bootStrapSampleData, bootStrapSampleTargets = bootStraps[strap]['boot']
                outOfBagSampleData, outOfBagSampleTargets = bootStraps[strap]['test']
            else:
                bootStrapSampleData, bootStrapSampleTargets = data, targets

            # boosting
            if self._boosting:
                self._booster.train(tree, bootStrapSampleData, bootStrapSampleTargets)
            else:
                # fit a decision tree model to the current sample
                tree.train(bootStrapSampleData, bootStrapSampleTargets)
            if i == 0 and self.retrainFirst is True:
                print('·'*75)
            # store the out-of-bag test set for the current bootstrap
            if (bootStraps[strap]['test'][0].size > 0) or (bootStraps[strap]['test'][1].size > 0):
                outOfBag[strap] = bootStraps[strap]['test']
            else:
                outOfBag[strap] = np.array([[],[]])
        if self.retrainFirst is True:
            self.trees = self.trees[1:]
            DecisionTree.treeID -= 1

    def eval(self, data: ArrayLike):
        """
        predict from the ensemble
        """
        # loop through each fitted model
        predictions = []
        for m in self.trees:
            # make predictions on the input data
            pred = m.eval(data)
            # append predictions to storage list
            predictions.append(pred.reshape(-1,1))

        # compute the ensemble prediction
        predictions = np.concatenate(predictions, axis=1)
        prediction = self.voting(predictions)

        # return the prediction
        return prediction

    def __str__(self):
        """
        used for printing the forrest in a human readable manner
        """
        treeStrings = [str(tree) for tree in self.trees]

        componentString = f'voting: {self.voting.name}, booster: {self._booster.name}, bootstrapping: {self.bootstrapping}\n'
        printString = ' forrest '.center(len(componentString), '━')
        printString += '\n'
        printString += componentString
        printString += '\n'
        for i, treeString in enumerate(treeStrings):
            printString += treeString
            if i < len(self.trees) + 1:
                printString += '\n\n'
        return printString

    @property
    def featureImportance(self) -> np.ndarray:
        """
        Computes the feature importance for each feature
        """
        featImportance = np.zeros(self.numFeatures)

        for tree in self.trees:
            featImportance += tree.featureImportance

        # not really sure I can do this, but wanted to clip out negative values
        featImportance = np.clip(featImportance, 0, None)
        return featImportance / sum(featImportance)

    def accuracy(self, data: np.ndarray, targets: np.ndarray) -> [namedtuple]:
        accuracy = namedtuple('Accuracy', 'name accuracy')
        accuarcies = []
        for tree in self.trees:
            score = accuracy(f'tree: {tree.id}', tree.accuracy(data, targets))
            accuarcies.append(score)
        return accuarcies

    def __len__(self) -> int:
        return len(self.trees)